@@ -73,14 +73,13 @@
 #if DEBUG_GLYPHS
 #undef DBG
 #define DBG(x) ErrorF x
-#else
-#define NDEBUG 1
 #endif
 
 #define FALLBACK 0
 #define NO_GLYPH_CACHE 0
 #define NO_GLYPHS_TO_DST 0
 #define NO_GLYPHS_VIA_MASK 0
+#define NO_SMALL_MASK 0
 #define NO_GLYPHS_SLOW 0
 
 #define CACHE_PICTURE_SIZE 1024
@@ -108,17 +107,9 @@ static void _assert_pixmap_contains_box(PixmapPtr pixmap, BoxPtr box, const char
 #define assert_pixmap_contains_box(p, b)
 #endif
 
-struct sna_glyph {
-	PicturePtr atlas;
-	struct sna_coordinate coordinate;
-	uint16_t size, pos;
-};
-
-static DevPrivateKeyRec sna_glyph_key;
-
-static inline struct sna_glyph *glyph_get_private(GlyphPtr glyph)
+static inline struct sna_glyph *sna_glyph(GlyphPtr glyph)
 {
-	return dixGetPrivateAddr(&glyph->devPrivates, &sna_glyph_key);
+	return (struct sna_glyph *)glyph->devPrivates;
 }
 
 #define NeedsComponent(f) (PICT_FORMAT_A(f) != 0 && PICT_FORMAT_RGB(f) != 0)
@@ -161,14 +152,20 @@ static Bool realize_glyph_caches(struct sna *sna)
 
 	DBG(("%s\n", __FUNCTION__));
 
+	if (sna->kgem.wedged || !sna->have_render)
+		return TRUE;
+
 	for (i = 0; i < ARRAY_SIZE(formats); i++) {
 		struct sna_glyph_cache *cache = &sna->render.glyph[i];
+		struct sna_pixmap *priv;
 		PixmapPtr pixmap;
-		PicturePtr picture;
+		PicturePtr picture = NULL;
+		PictFormatPtr pPictFormat;
 		CARD32 component_alpha;
 		int depth = PIXMAN_FORMAT_DEPTH(formats[i]);
 		int error;
-		PictFormatPtr pPictFormat = PictureMatchFormat(screen, depth, formats[i]);
+
+		pPictFormat = PictureMatchFormat(screen, depth, formats[i]);
 		if (!pPictFormat)
 			goto bail;
 
@@ -181,13 +178,18 @@ static Bool realize_glyph_caches(struct sna *sna)
 		if (!pixmap)
 			goto bail;
 
-		component_alpha = NeedsComponent(pPictFormat->format);
-		picture = CreatePicture(0, &pixmap->drawable, pPictFormat,
-					CPComponentAlpha, &component_alpha,
-					serverClient, &error);
+		priv = sna_pixmap(pixmap);
+		if (priv != NULL) {
+			/* Prevent the cache from ever being paged out */
+			priv->pinned = true;
+
+			component_alpha = NeedsComponent(pPictFormat->format);
+			picture = CreatePicture(0, &pixmap->drawable, pPictFormat,
+						CPComponentAlpha, &component_alpha,
+						serverClient, &error);
+		}
 
 		screen->DestroyPixmap(pixmap);
-
 		if (!picture)
 			goto bail;
 
@@ -310,7 +312,12 @@ glyph_cache(ScreenPtr screen,
 
 	if (glyph->info.width > GLYPH_MAX_SIZE ||
 	    glyph->info.height > GLYPH_MAX_SIZE) {
-		((PixmapPtr)glyph_picture->pDrawable)->usage_hint = 0;
+		PixmapPtr pixmap = (PixmapPtr)glyph_picture->pDrawable;
+		assert(glyph_picture->pDrawable->type == DRAWABLE_PIXMAP);
+		if (pixmap->drawable.depth >= 8) {
+			pixmap->usage_hint = SNA_CREATE_GLYPH;
+			sna_pixmap_force_to_gpu(pixmap, MOVE_READ);
+		}
 		return FALSE;
 	}
 
@@ -356,7 +363,7 @@ glyph_cache(ScreenPtr screen,
 	}
 	assert(cache->glyphs[pos] == NULL);
 
-	priv = glyph_get_private(glyph);
+	priv = sna_glyph(glyph);
 	cache->glyphs[pos] = priv;
 	priv->atlas = cache->picture;
 	priv->size = size;
@@ -395,7 +402,6 @@ static void apply_damage(struct sna_composite_op *op,
 	sna_damage_add_box(op->damage, &box);
 }
 
-#define GET_PRIVATE(g) ((struct sna_glyph *)((char *)(g)->devPrivates + priv_offset))
 static Bool
 glyphs_to_dst(struct sna *sna,
 	      CARD8 op,
@@ -406,7 +412,6 @@ glyphs_to_dst(struct sna *sna,
 {
 	struct sna_composite_op tmp;
 	ScreenPtr screen = dst->pDrawable->pScreen;
-	const int priv_offset = sna_glyph_key.offset;
 	int index = screen->myNum;
 	PicturePtr glyph_atlas;
 	BoxPtr rects;
@@ -443,7 +448,7 @@ glyphs_to_dst(struct sna *sna,
 			if (glyph->info.width == 0 || glyph->info.height == 0)
 				goto next_glyph;
 
-			priv = *GET_PRIVATE(glyph);
+			priv = *sna_glyph(glyph);
 			if (priv.atlas == NULL) {
 				if (glyph_atlas) {
 					tmp.done(sna, &tmp);
@@ -454,7 +459,7 @@ glyphs_to_dst(struct sna *sna,
 					priv.atlas = GlyphPicture(glyph)[index];
 					priv.coordinate.x = priv.coordinate.y = 0;
 				} else
-					priv = *GET_PRIVATE(glyph);
+					priv = *sna_glyph(glyph);
 			}
 
 			if (priv.atlas != glyph_atlas) {
@@ -536,7 +541,6 @@ glyphs_slow(struct sna *sna,
 {
 	struct sna_composite_op tmp;
 	ScreenPtr screen = dst->pDrawable->pScreen;
-	const int priv_offset = sna_glyph_key.offset;
 	int index = screen->myNum;
 	int16_t x, y;
 
@@ -567,14 +571,14 @@ glyphs_slow(struct sna *sna,
 			if (glyph->info.width == 0 || glyph->info.height == 0)
 				goto next_glyph;
 
-			priv = *GET_PRIVATE(glyph);
+			priv = *sna_glyph(glyph);
 			if (priv.atlas == NULL) {
 				if (!glyph_cache(screen, &sna->render, glyph)) {
 					/* no cache for this glyph */
 					priv.atlas = GlyphPicture(glyph)[index];
 					priv.coordinate.x = priv.coordinate.y = 0;
 				} else
-					priv = *GET_PRIVATE(glyph);
+					priv = *sna_glyph(glyph);
 			}
 
 			DBG(("%s: glyph=(%d, %d)x(%d, %d), src=(%d, %d), mask=(%d, %d)\n",
@@ -651,16 +655,11 @@ next_glyph:
 	return TRUE;
 }
 
-static void
+static bool
 clear_pixmap(struct sna *sna, PixmapPtr pixmap)
 {
 	struct sna_pixmap *priv = sna_pixmap(pixmap);
-	assert(priv->gpu_only);
-	sna->render.fill_one(sna, pixmap, priv->gpu_bo, 0,
-			     0, 0,
-			     pixmap->drawable.width,
-			     pixmap->drawable.height,
-			     GXclear);
+	return sna->render.clear(sna, pixmap, priv->gpu_bo);
 }
 
 static Bool
@@ -674,7 +673,6 @@ glyphs_via_mask(struct sna *sna,
 {
 	ScreenPtr screen = dst->pDrawable->pScreen;
 	struct sna_composite_op tmp;
-	const int priv_offset = sna_glyph_key.offset;
 	int index = screen->myNum;
 	CARD32 component_alpha;
 	PixmapPtr pixmap;
@@ -724,94 +722,202 @@ glyphs_via_mask(struct sna *sna,
 			return FALSE;
 	}
 
-	pixmap = screen->CreatePixmap(screen,
-				      width, height, format->depth,
-				      SNA_CREATE_SCRATCH);
-	if (!pixmap)
-		return FALSE;
-
 	component_alpha = NeedsComponent(format->format);
-	mask = CreatePicture(0, &pixmap->drawable,
-			     format, CPComponentAlpha,
-			     &component_alpha, serverClient, &error);
-	screen->DestroyPixmap(pixmap);
-	if (!mask)
-		return FALSE;
+	if (!NO_SMALL_MASK &&
+	    (uint32_t)width * height * format->depth < 8 * 4096) {
+		pixman_image_t *mask_image;
+		int s;
 
-	ValidatePicture(mask);
-	clear_pixmap(sna, pixmap);
+		DBG(("%s: small mask [format=%lx, depth=%d, size=%d], rendering glyphs to upload buffer\n",
+		     __FUNCTION__, (unsigned long)format->format,
+		     format->depth, (uint32_t)width*height*format->depth));
 
-	memset(&tmp, 0, sizeof(tmp));
-	glyph_atlas = NULL;
-	do {
-		int n = list->len;
-		x += list->xOff;
-		y += list->yOff;
-		while (n--) {
-			GlyphPtr glyph = *glyphs++;
-			struct sna_glyph *priv;
-			PicturePtr this_atlas;
-			struct sna_composite_rectangles r;
+upload:
+		pixmap = sna_pixmap_create_upload(screen,
+						  width, height,
+						  format->depth,
+						  KGEM_BUFFER_WRITE);
+		if (!pixmap)
+			return FALSE;
 
-			if (glyph->info.width == 0 || glyph->info.height == 0)
-				goto next_glyph;
+		mask_image =
+			pixman_image_create_bits(format->depth << 24 | format->format,
+						 width, height,
+						 pixmap->devPrivate.ptr,
+						 pixmap->devKind);
+		if (mask_image == NULL) {
+			screen->DestroyPixmap(pixmap);
+			return FALSE;
+		}
 
-			priv = GET_PRIVATE(glyph);
-			if (priv->atlas != NULL) {
-				this_atlas = priv->atlas;
-				r.src = priv->coordinate;
-			} else {
-				if (glyph_atlas) {
-					tmp.done(sna, &tmp);
-					glyph_atlas = NULL;
+		memset(pixmap->devPrivate.ptr, 0, pixmap->devKind*height);
+		s = dst->pDrawable->pScreen->myNum;
+		do {
+			int n = list->len;
+			x += list->xOff;
+			y += list->yOff;
+			while (n--) {
+				GlyphPtr g = *glyphs++;
+				PicturePtr picture;
+				pixman_image_t *glyph_image;
+				int16_t xi, yi;
+
+				if (g->info.width == 0 || g->info.height == 0)
+					goto next_image;
+
+				/* If the mask has been cropped, it is likely
+				 * that some of the glyphs fall outside.
+				 */
+				xi = x - g->info.x;
+				yi = y - g->info.y;
+				if (xi >= width || yi >= height)
+					goto next_image;
+				if (xi + g->info.width  <= 0 ||
+				    yi + g->info.height <= 0)
+					goto next_image;
+
+				glyph_image = sna_glyph(g)->image;
+				if (glyph_image == NULL) {
+					int dx, dy;
+
+					picture = GlyphPicture(g)[s];
+					if (picture == NULL)
+						goto next_image;
+
+					glyph_image = image_from_pict(picture,
+								      FALSE,
+								      &dx, &dy);
+					if (!glyph_image)
+						goto next_image;
+
+					assert(dx == 0 && dy == 0);
+					sna_glyph(g)->image = glyph_image;
 				}
-				if (glyph_cache(screen, &sna->render, glyph)) {
+
+				DBG(("%s: glyph to mask (%d, %d)x(%d, %d)\n",
+				     __FUNCTION__,
+				     xi, yi,
+				     g->info.width,
+				     g->info.height));
+
+				pixman_image_composite(PictOpAdd,
+						       glyph_image,
+						       NULL,
+						       mask_image,
+						       0, 0,
+						       0, 0,
+						       xi, yi,
+						       g->info.width,
+						       g->info.height);
+
+next_image:
+				x += g->info.xOff;
+				y += g->info.yOff;
+			}
+			list++;
+		} while (--nlist);
+		pixman_image_unref(mask_image);
+
+		mask = CreatePicture(0, &pixmap->drawable,
+				     format, CPComponentAlpha,
+				     &component_alpha, serverClient, &error);
+		screen->DestroyPixmap(pixmap);
+		if (!mask)
+			return FALSE;
+
+		ValidatePicture(mask);
+	} else {
+		pixmap = screen->CreatePixmap(screen,
+					      width, height, format->depth,
+					      SNA_CREATE_SCRATCH);
+		if (!pixmap)
+			return FALSE;
+
+		mask = CreatePicture(0, &pixmap->drawable,
+				     format, CPComponentAlpha,
+				     &component_alpha, serverClient, &error);
+		screen->DestroyPixmap(pixmap);
+		if (!mask)
+			return FALSE;
+
+		ValidatePicture(mask);
+		if (!clear_pixmap(sna, pixmap)) {
+			FreePicture(mask, 0);
+			goto upload;
+		}
+
+		memset(&tmp, 0, sizeof(tmp));
+		glyph_atlas = NULL;
+		do {
+			int n = list->len;
+			x += list->xOff;
+			y += list->yOff;
+			while (n--) {
+				GlyphPtr glyph = *glyphs++;
+				struct sna_glyph *priv;
+				PicturePtr this_atlas;
+				struct sna_composite_rectangles r;
+
+				if (glyph->info.width == 0 || glyph->info.height == 0)
+					goto next_glyph;
+
+				priv = sna_glyph(glyph);
+				if (priv->atlas != NULL) {
 					this_atlas = priv->atlas;
 					r.src = priv->coordinate;
 				} else {
-					/* no cache for this glyph */
-					this_atlas = GlyphPicture(glyph)[index];
-					r.src.x = r.src.y = 0;
-				}
-			}
-
-			if (this_atlas != glyph_atlas) {
-				if (glyph_atlas)
-					tmp.done(sna, &tmp);
-
-				if (!sna->render.composite(sna, PictOpAdd,
-							   this_atlas, NULL, mask,
-							   0, 0, 0, 0, 0, 0,
-							   0, 0,
-							   &tmp)) {
-					FreePicture(mask, 0);
-					return FALSE;
+					if (glyph_atlas) {
+						tmp.done(sna, &tmp);
+						glyph_atlas = NULL;
+					}
+					if (glyph_cache(screen, &sna->render, glyph)) {
+						this_atlas = priv->atlas;
+						r.src = priv->coordinate;
+					} else {
+						/* no cache for this glyph */
+						this_atlas = GlyphPicture(glyph)[index];
+						r.src.x = r.src.y = 0;
+					}
 				}
 
-				glyph_atlas = this_atlas;
-			}
+				if (this_atlas != glyph_atlas) {
+					if (glyph_atlas)
+						tmp.done(sna, &tmp);
 
-			DBG(("%s: blt glyph origin (%d, %d), offset (%d, %d), src (%d, %d), size (%d, %d)\n",
-			     __FUNCTION__,
-			     x, y,
-			     glyph->info.x, glyph->info.y,
-			     r.src.x, r.src.y,
-			     glyph->info.width, glyph->info.height));
+					if (!sna->render.composite(sna, PictOpAdd,
+								   this_atlas, NULL, mask,
+								   0, 0, 0, 0, 0, 0,
+								   0, 0,
+								   &tmp)) {
+						FreePicture(mask, 0);
+						return FALSE;
+					}
 
-			r.dst.x = x - glyph->info.x;
-			r.dst.y = y - glyph->info.y;
-			r.width  = glyph->info.width;
-			r.height = glyph->info.height;
-			tmp.blt(sna, &tmp, &r);
+					glyph_atlas = this_atlas;
+				}
+
+				DBG(("%s: blt glyph origin (%d, %d), offset (%d, %d), src (%d, %d), size (%d, %d)\n",
+				     __FUNCTION__,
+				     x, y,
+				     glyph->info.x, glyph->info.y,
+				     r.src.x, r.src.y,
+				     glyph->info.width, glyph->info.height));
+
+				r.dst.x = x - glyph->info.x;
+				r.dst.y = y - glyph->info.y;
+				r.width  = glyph->info.width;
+				r.height = glyph->info.height;
+				tmp.blt(sna, &tmp, &r);
 
 next_glyph:
-			x += glyph->info.xOff;
-			y += glyph->info.yOff;
-		}
-		list++;
-	} while (--nlist);
-	if (glyph_atlas)
-		tmp.done(sna, &tmp);
+				x += glyph->info.xOff;
+				y += glyph->info.yOff;
+			}
+			list++;
+		} while (--nlist);
+		if (glyph_atlas)
+			tmp.done(sna, &tmp);
+	}
 
 	sna_composite(op,
 		      src, mask, dst,
@@ -824,17 +930,6 @@ next_glyph:
 	return TRUE;
 }
 
-Bool sna_glyphs_init(ScreenPtr screen)
-{
-	if (!dixRegisterPrivateKey(&sna_glyph_key,
-				   PRIVATE_GLYPH,
-				   sizeof(struct sna_glyph)))
-		return FALSE;
-
-	return TRUE;
-	(void)screen;
-}
-
 Bool sna_glyphs_create(struct sna *sna)
 {
 	return realize_glyph_caches(sna);
@@ -931,7 +1026,7 @@ glyphs_fallback(CARD8 op,
 	DBG(("%s: (%d, %d), (%d, %d)\n",
 	     __FUNCTION__, box.x1, box.y1, box.x2, box.y2));
 
-	RegionInit(&region, &box, 1);
+	RegionInit(&region, &box, 0);
 	RegionTranslate(&region, dst->pDrawable->x, dst->pDrawable->y);
 	if (dst->pCompositeClip)
 		RegionIntersect(&region, &region, dst->pCompositeClip);
@@ -942,9 +1037,24 @@ glyphs_fallback(CARD8 op,
 	if (!RegionNotEmpty(&region))
 		return;
 
-	sna_drawable_move_region_to_cpu(dst->pDrawable, &region, true);
-	if (src->pDrawable)
-		sna_drawable_move_to_cpu(src->pDrawable, false);
+	if (!sna_drawable_move_region_to_cpu(dst->pDrawable, &region,
+					     MOVE_READ | MOVE_WRITE))
+		return;
+	if (dst->alphaMap &&
+	    !sna_drawable_move_to_cpu(dst->alphaMap->pDrawable,
+				      MOVE_READ | MOVE_WRITE))
+		return;
+
+	if (src->pDrawable) {
+		if (!sna_drawable_move_to_cpu(src->pDrawable,
+					      MOVE_READ))
+			return;
+
+		if (src->alphaMap &&
+		    !sna_drawable_move_to_cpu(src->alphaMap->pDrawable,
+					      MOVE_READ))
+			return;
+	}
 	RegionTranslate(&region, -dst->pDrawable->x, -dst->pDrawable->y);
 
 	dst_image = image_from_pict(dst, TRUE, &x, &y);
@@ -1001,25 +1111,34 @@ glyphs_fallback(CARD8 op,
 		y += list->yOff;
 		while (n--) {
 			GlyphPtr g = *glyphs++;
-			PicturePtr picture;
 			pixman_image_t *glyph_image;
-			int dx, dy;
 
 			if (g->info.width == 0 || g->info.height == 0)
 				goto next_glyph;
 
-			picture = GlyphPicture(g)[screen];
-			if (picture == NULL)
-				goto next_glyph;
+			glyph_image = sna_glyph(g)->image;
+			if (glyph_image == NULL) {
+				PicturePtr picture;
+				int dx, dy;
 
-			glyph_image = image_from_pict(picture, FALSE, &dx, &dy);
-			if (!glyph_image)
-				goto next_glyph;
+				picture = GlyphPicture(g)[screen];
+				if (picture == NULL)
+					goto next_glyph;
+
+				glyph_image = image_from_pict(picture,
+							      FALSE,
+							      &dx, &dy);
+				if (!glyph_image)
+					goto next_glyph;
+
+				assert(dx == 0 && dy == 0);
+				sna_glyph(g)->image = glyph_image;
+			}
 
 			if (mask_format) {
 				DBG(("%s: glyph+(%d,%d) to mask (%d, %d)x(%d, %d)\n",
 				     __FUNCTION__,
-				     dx,dy,
+				     dx, dy,
 				     x - g->info.x,
 				     y - g->info.y,
 				     g->info.width,
@@ -1036,28 +1155,29 @@ glyphs_fallback(CARD8 op,
 						       g->info.width,
 						       g->info.height);
 			} else {
-				DBG(("%s: glyph+(%d, %d) to dst (%d, %d)x(%d, %d)\n",
+				int xi = x - g->info.x;
+				int yi = y - g->info.y;
+
+				DBG(("%s: glyph+(%d, %d) to dst (%d, %d)x(%d, %d), src (%d, %d) [op=%d]\n",
 				     __FUNCTION__,
 				     dx, dy,
-				     x - g->info.x,
-				     y - g->info.y,
-				     g->info.width,
-				     g->info.height));
+				     xi, yi,
+				     g->info.width, g->info.height,
+				     src_x + xi,
+				     src_y + yi,
+				     op));
 
 				pixman_image_composite(op,
 						       src_image,
 						       glyph_image,
 						       dst_image,
-						       src_x + (x - g->info.x),
-						       src_y + (y - g->info.y),
+						       src_x + xi,
+						       src_y + yi,
 						       dx, dy,
-						       x - g->info.x,
-						       y - g->info.y,
+						       xi, yi,
 						       g->info.width,
 						       g->info.height);
 			}
-			free_pixman_pict(picture,glyph_image);
-
 next_glyph:
 			x += g->info.xOff;
 			y += g->info.yOff;
@@ -1106,17 +1226,11 @@ sna_glyphs(CARD8 op,
 	if (REGION_NUM_RECTS(dst->pCompositeClip) == 0)
 		return;
 
-	if (FALLBACK)
+	if (FALLBACK || DEBUG_NO_RENDER)
 		goto fallback;
 
-	if (sna->kgem.wedged || !sna->have_render) {
-		DBG(("%s: no render (wedged=%d)\n",
-		     __FUNCTION__, sna->kgem.wedged));
-		goto fallback;
-	}
-
-	if (!is_gpu(dst->pDrawable)) {
-		DBG(("%s: fallback -- no destination bo\n", __FUNCTION__));
+	if (wedged(sna)) {
+		DBG(("%s: wedged\n", __FUNCTION__));
 		goto fallback;
 	}
 
@@ -1126,14 +1240,17 @@ sna_glyphs(CARD8 op,
 		goto fallback;
 	}
 
-	if (dst->alphaMap || src->alphaMap) {
-		DBG(("%s: fallback -- alpha maps\n", __FUNCTION__));
+	if (dst->alphaMap) {
+		DBG(("%s: fallback -- dst alpha map\n", __FUNCTION__));
 		goto fallback;
 	}
 
+	_mask = mask;
 	/* XXX discard the mask for non-overlapping glyphs? */
 
-	if (!mask || (op == PictOpAdd && dst->format == mask->format)) {
+	if (!_mask ||
+	    (((nlist == 1 && list->len == 1) || op == PictOpAdd) &&
+	     dst->format == (_mask->depth << 24 | _mask->format))) {
 		if (glyphs_to_dst(sna, op,
 				  src, dst,
 				  src_x, src_y,
@@ -1141,7 +1258,6 @@ sna_glyphs(CARD8 op,
 			return;
 	}
 
-	_mask = mask;
 	if (!_mask)
 		_mask = glyphs_format(nlist, list, glyphs);
 	if (_mask) {
@@ -1165,19 +1281,20 @@ fallback:
 void
 sna_glyph_unrealize(ScreenPtr screen, GlyphPtr glyph)
 {
-	struct sna_glyph_cache *cache;
-	struct sna_glyph *priv;
-	struct sna *sna;
+	struct sna_glyph *priv = sna_glyph(glyph);
 
-	priv = glyph_get_private(glyph);
-	if (priv->atlas == NULL)
-		return;
+	if (priv->image) {
+		pixman_image_unref(priv->image);
+		priv->image = NULL;
+	}
 
-	sna = to_sna_from_screen(screen);
-	cache = &sna->render.glyph[priv->pos & 1];
-	assert(cache->glyphs[priv->pos >> 1] == priv);
-	cache->glyphs[priv->pos >> 1] = NULL;
-	priv->atlas = NULL;
+	if (priv->atlas) {
+		struct sna *sna = to_sna_from_screen(screen);
+		struct sna_glyph_cache *cache = &sna->render.glyph[priv->pos&1];
+		assert(cache->glyphs[priv->pos >> 1] == priv);
+		cache->glyphs[priv->pos >> 1] = NULL;
+		priv->atlas = NULL;
+	}
 }
 
 void sna_glyphs_close(struct sna *sna)
